/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <assert.h>
#include <string.h>
#include "libs.h"
#include "libs-package.h"
#include "os_dep.h"
#include "error.h"
#include "threads.h"

/**
 *  Search for library file in dir_list. This routine doesn't copy paths out
 *  from the list, but replaces separators with \0. This means no extra buffers,
 *  no memcpy/strcpy, but dir_list must be writeable. At the end, dir_list
 *  is restored.
 *
 * @param library_name The library to find
 * @param file_name Filename of the library (path relative from current)
 * @param dep_file_name Filename of the .dep file (path relative from current)
 */
int gpmi_library_find_path(char *dir_list, const char *library_name, char *file_name, char *dep_file_name, int path_max)
{
	char *start, *end;
	int at_end, fd;

	at_end = 0;
	for (start = dir_list, end = dir_list; ; end++) {
		switch (*end) {
			case '\0': /* End of string */
				at_end = 1;
			case GPMI_PATH_SEP_CHAR:  /* Separator */
				/* Assamble paths */
				*end = '\0';
				gpmi_library_add_path(start, library_name, file_name, dep_file_name, path_max);

				/* Check if lib file exists */
				fd = open(file_name, O_RDONLY);
				if (fd >= 0) {
					/* Restore separator */
					if (!at_end)
						*end  = GPMI_PATH_SEP_CHAR;

					close(fd);
					return 0;
				}

				/* If we are at the end, exit */
				if (at_end) {
					gpmi_err_stack_push(gpmi_err_file_not_found, gpmi_err_src_find_path, file_name);
					return gpmi_err_file_not_found;
				}

				/* Restore separator, prepare next token */
				*end  = GPMI_PATH_SEP_CHAR;
				start = end + 1;
		}
	}

	/* We can't get here, but let's not end an int function without a return */
	return gpmi_invalid_int;
}

/**
 * Lowlevel function to load a dynamic library.
 *
 * @param state If not NULL, the content can be: 0: lib already exists, 1: lib loaded, other: error
 */
gpmi_library *gpmi_load_library(char *dir, const char *library_name, int *state, int is_pkg)
{
	char file_name[PATH_MAX];
	char dep_file_name[PATH_MAX];
	gpmi_library *library;
	int depstat, ret;

#ifndef gpmi_no_static
	/* Check wether the library is a static lib */
	snprintf(file_name, sizeof(file_name), "__STATIC__%s", library_name);
	library = gpmi_find_library(file_name);

	/* Return if found */
	if (library != NULL) {
		/* Increase the reference counter */
		library->references++;
		if (state != NULL)
			*state = 0;

		return library;
	}
#endif

	/* Find library using the search path; if found, file_name and dep_file_name are set */
	ret = gpmi_library_find_path(dir, library_name, file_name, dep_file_name, sizeof(file_name));
	if (ret != 0) {
		if (state != NULL)
			*state = ret;
		return NULL;
	}

	/* Check if we already have this library loaded */
	library = gpmi_find_library(file_name);

	if (library != NULL) {
		/* Avoid loading ourself inside ourself ;) */
		if ((library->flags & gpmi_libf_loading) == gpmi_libf_loading) {
			if (state != NULL)
				*state = gpmi_err_already_loading;
			gpmi_err_stack_push(gpmi_err_already_loading, gpmi_err_src_load_library, library->name);
			return NULL;
		}

		/* Increase the reference counter */
		library->references++;

		/* Load dependencies */
		depstat = gpmi_libdep_load(library, dep_file_name);

		/* Return if it fails */
		if (depstat < 0) {
			if (state != NULL)
				*state = depstat;
			gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_load_library, library->name);
			return NULL;
		}

		return library;
	}

	/* Mark as new library */
	if (state != NULL)
		*state = 1;

	library = gpmi_lib_create(file_name);
	library->flags = gpmi_libf_loading;

	/* First load dependencies */
	depstat = gpmi_libdep_load(library, dep_file_name);

	/* Return if it fails */
	if (depstat < 0) {
		ret = depstat;
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_load_library, library->name);
		goto free_lib_and_return_null;
	}

	/* Try to load the library */
	ret = gpmi_library_load_lowlevel(library, file_name);
	if (ret) {
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_load_library, file_name);
		goto free_lib_and_return_null;
	}

	/* We have now one reference */
	library->references = 1;

	/* Library is loaded now */
	library->flags &= ~gpmi_libf_loading;

	return library;


free_lib_and_return_null: ;
	free(library->name);
	free(library);
	if (state != NULL)
		*state = ret;
	return NULL;
}

/**
 * Decrease reference number of a library and unload it when
 * it's not needed anymore. Call uninit() if the library is being unloaded.
 */
void gpmi_unload_library(gpmi_library *library, void (*uninit)() )
{
	gpmi_library *tmp;

	/* Unload the deps */
	gpmi_libdep_unload(library);

	/* Decrease the reference counter */
	library->references--;

	/* If it hits below 0, something went pretty wrong (an unload without a load) */
	assert(library->references >= 0);

	/* There are still links to this lib, don't unload */
	if (library->references > 0)
		return;

	/* Run uninit, both in static and dynamic mode */
	if (uninit != NULL)
		uninit();

	/* Free the library name */
	if (library->name != NULL)
		free(library->name);

#ifndef gpmi_no_static
	/* Some things we don't want to do on static libraries */
	if ((library->flags & gpmi_libf_static) == gpmi_libf_static) {
	} else
#endif
	{
		/* Unload the library */
		gpmi_library_unload_lowlevel(library);
	}

	gpmi_shared_lock;
	/* Remove the library pointer from the linked list of libraries */
	for (tmp = gpmi_shared.gpmi_libraries; tmp != NULL; tmp = tmp->next)
	{
		if (tmp == library) {
			/* Remove the library from the list */
			if (tmp->prev != NULL)
				tmp->prev->next = tmp->next;
			else
				gpmi_shared.gpmi_libraries  = tmp->next;

			if (tmp->next != NULL)
				tmp->next->prev = tmp->prev;

			break;
		}
	}
	gpmi_shared_unlock;

	/* Free the library */
	free(library);
}

/**
 *  Find a library by name in a linked list. Case sensitive.
 */
gpmi_library *gpmi_find_library(const char *library_name)
{
	struct gpmi_library *library;

	gpmi_shared_lock;

	/* Check if library list is empty */
	if (gpmi_shared.gpmi_libraries == NULL) {
		gpmi_shared_unlock;
		return NULL;
	}

	/* Search name for the linked list */
	for (library = gpmi_shared.gpmi_libraries; library != NULL; library = library->next) {
		if ((library->name != NULL) && (strcmp(library_name, library->name) == 0)) {
			gpmi_shared_unlock;
			return library;
		}
	}

	/* Not found */
	gpmi_shared_unlock;
	return NULL;
}

/**
 * Creates a library and adds it to the global library list.
 */
gpmi_library *gpmi_lib_create(char *library_name)
{
	gpmi_library *library;

	/* Alloc and add new library structure */
	library = malloc(sizeof(gpmi_library));
	memset(library, 0, sizeof(gpmi_library));

	gpmi_shared_lock;
	/* First library? */
	if (gpmi_shared.gpmi_libraries == NULL) {
		gpmi_shared.gpmi_libraries = library;
	} else {
		/* Insert new lib at the beginning of the linked list */
		gpmi_shared.gpmi_libraries->prev = library;
		library->next                    = gpmi_shared.gpmi_libraries;
		gpmi_shared.gpmi_libraries       = library;
	}

	gpmi_shared_unlock;
	library->name = strdup(library_name);

	return library;
}

/**
 *  Load dependency of a library recursively.
 */
int gpmi_libdep_load(gpmi_library *lib, const char *dep_file_name)
{
	char *s, buff[PATH_MAX], *vers;
	int c, len, state, veri;
	FILE *f;
	gpmi_package *pkg;

	/* try to open the dep file of the library */
	f = fopen(dep_file_name, "r");
	if (f == NULL)
		return 0;

	s   = buff;
	*s  = 0;
	len = 0;

	/* Read the file char by char seeking words */
	do {
		c = fgetc(f);
		switch(c) {
			/* end of word */
			case EOF:
			case ',':
			case ' ':
			case ';':
			case '\n':
			case '\r':
			case '\t':
				*s = '\0';

				if (*buff != 0) {
					/* Split version number by replacing the first ':' with '\0' */
					vers = strchr(buff, ':');
					if (vers != NULL) {
						*vers = 0;
						/* Atoi will stop at the next ':' - later we should use strtol here
						   if we are interested in flags after the 2nd : */
						veri = atoi(vers + 1);
					}
					else
						veri = 0;

					/* Try to load the package */
					state = gpmi_pkg_load(buff, veri, NULL, NULL, &pkg);

					/* Don't go in endless recursions becuase cyclic deps */
					if ((state == gpmi_err_already_loading) || (state == gpmi_err_cyclic_dep)) {
						fclose(f);
						gpmi_err_stack_push(gpmi_err_cyclic_dep, gpmi_err_src_libdep_load, buff);
						return gpmi_err_cyclic_dep;
					}

					/* Check for fatal errors */
					if (state < 0) {
						fclose(f);
						return state;
					}

					/* XXX: we need a flag, strncmp is ugly! */
					if (strncmp(pkg->library->name, "__STATIC__", 10) != 0) {
						/* Add the package handle in the list of dep-loaded packages;
						   static packages are not loaded, we won't need to unload them so
							 don't add those. */
						pkg->next   = lib->libdep;
						lib->libdep = pkg;
					}
				}

				s   = buff;
				*s  = 0;
				len = 0;
				break;

			/* copy next character */
			default:
				*s = c;
				s++;
				len++;
				if (len >= sizeof(buff)) {
					fclose(f);
					gpmi_err_stack_push(gpmi_err_too_long, gpmi_err_src_libdep_load, NULL);
					return gpmi_err_too_long;
				}
		}
	} while(c != EOF);

	fclose(f);
	return 0;
}

/**
 *  Unload dependency of a library recursively.
 */
void gpmi_libdep_unload(gpmi_library *lib)
{
	gpmi_package *pkg;

	/* Walk through all the deps and free them */
	while ((pkg = lib->libdep) != NULL) {
		lib->libdep = pkg->next;
		gpmi_pkg_unload(pkg);
	}

	lib->libdep = NULL;
}
