/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & glx & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "libs.h"
#include "os_dep.h"
#include "error.h"

#ifdef WIN32
#include <windows.h>
#include <shlobj.h>
#endif

char *osdep_GPMI_INSTALL_PATH_USER_MOD_LIB = NULL;
char *osdep_GPMI_INSTALL_PATH_USER_PKG_LIB = NULL;
char *osdep_GPMI_INSTALL_PATH_MOD_LIB = NULL;
char *osdep_GPMI_INSTALL_PATH_PKG_LIB = NULL;

#ifdef WIN32
char *gpmi_registry_get_path(char *name)
{
	HKEY hKey;
	LONG lRet;
	char buf[PATH_MAX];
	DWORD size;

	// check HKCU\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_CURRENT_USER, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = sizeof(buf);
		lRet = RegQueryValueEx(hKey, name, NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS)
			return strdup(buf);
	}

	// check HKLM\SOFTWARE\LibGPMI\InstallPath
	lRet = RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SOFTWARE\\LibGPMI", 0, KEY_QUERY_VALUE, &hKey);
	if (lRet == ERROR_SUCCESS) {
		size = sizeof(buf);
		lRet = RegQueryValueEx(hKey, name, NULL, NULL, (LPBYTE)buf, &size);
		RegCloseKey(hKey);
		if (lRet == ERROR_SUCCESS)
			return strdup(buf);
	}
	return NULL;
}
#endif


void gpmi_path_init()
{
	if (osdep_GPMI_INSTALL_PATH_USER_MOD_LIB != NULL)
		return;
	#ifdef WIN32
		/* On windows we try to get paths from the registry and fall back to compile-time values */
		#define pathdef(var, def) if (var == NULL) var = strdup(def);
		osdep_GPMI_INSTALL_PATH_USER_MOD_LIB = gpmi_registry_get_path("INSTALL_PATH_USER_MOD_LIB");
		osdep_GPMI_INSTALL_PATH_USER_PKG_LIB = gpmi_registry_get_path("INSTALL_PATH_USER_PKG_LIB");
		osdep_GPMI_INSTALL_PATH_MOD_LIB      = gpmi_registry_get_path("INSTALL_PATH_MOD_LIB");
		osdep_GPMI_INSTALL_PATH_PKG_LIB      = gpmi_registry_get_path("INSTALL_PATH_PKG_LIB");
	#else
		/* We have only the compile-time values */
		#define pathdef(var, def) var = strdup(def);
	#endif
	pathdef(osdep_GPMI_INSTALL_PATH_USER_MOD_LIB, GPMI_INSTALL_PATH_USER_MOD_LIB);
	pathdef(osdep_GPMI_INSTALL_PATH_USER_PKG_LIB, GPMI_INSTALL_PATH_USER_PKG_LIB);
	pathdef(osdep_GPMI_INSTALL_PATH_MOD_LIB,      GPMI_INSTALL_PATH_MOD_LIB);
	pathdef(osdep_GPMI_INSTALL_PATH_PKG_LIB,      GPMI_INSTALL_PATH_PKG_LIB);

	/* Just to make sure it won't mess up anything later */
	#undef pathdef
}

void gpmi_path_uninit()
{
	#define pathfree(var) if(var != NULL) free(var);

	pathfree(osdep_GPMI_INSTALL_PATH_USER_MOD_LIB);
	pathfree(osdep_GPMI_INSTALL_PATH_USER_PKG_LIB);
	pathfree(osdep_GPMI_INSTALL_PATH_MOD_LIB);
	pathfree(osdep_GPMI_INSTALL_PATH_PKG_LIB);

	#undef pathfree
}

#if defined(WIN32) && !defined(__CYGWIN__)
/* Function basename is not native on windows */
char *basename(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return path;
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) return (++last);
	return path;
}

/* Function dirname is not native on windows */
char *dirname(char *path)
{
	char* last = strrchr(path, '/');
	if ((path == NULL) || (*path == '\0')) return ".";
	if (*path == '/' && *(path + 1) == '\0') return "/";
	if (last == NULL) return ".";
	if (last && *(last + 1) == '\0') {
		*last = '\0';
		last = strrchr(path, '/');
	}
	if (last) *last = '\0';
	if (*path == '\0') return "/";
	return path;
}

#endif

/**
 * Build full pathname of a library and its dep
 */
void gpmi_library_add_path(const char *dir, const char *library_name, char *lib_file_name, char *dep_file_name, int path_size)
{

/* Os dependent part: the file name of a dynamic lib */
#ifdef DLCOMPAT
		/* With dlcompat, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.dlc",	dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#elif DARWIN
		/* On darwin, it is like unix, only so -> dylib */
		snprintf(lib_file_name, path_size,      "%s/%s.dylib",     dir, library_name);
		snprintf(dep_file_name, path_size,      "%s/%s.dep",    dir, library_name);
#else
		/* On unix, just add dir, file name and extension */
		snprintf(lib_file_name,	path_size,	"%s/%s.so",		dir, library_name);
		snprintf(dep_file_name,	path_size,	"%s/%s.dep",	dir, library_name);
#endif
}

/**
 *  Os dependent part for loading a dynamic linked library
 */
int gpmi_library_load_lowlevel(gpmi_library *library, char *file_name)
{
	char *error_str;
	library->handle = dlopen(file_name, RTLD_GLOBAL | RTLD_LAZY);
	if (library->handle == NULL) {
		error_str = malloc(256);
		snprintf(error_str, 256, "%s on file %s", dlerror(), file_name);
		gpmi_err_stack_push(gpmi_err_load_library, gpmi_err_src_lib_load_lowlevel, error_str);
		free(error_str);
		return 1;
	}

	return 0;
}

/**
 *  OS dependent part to unload a dynamic link library
 */
void gpmi_library_unload_lowlevel(gpmi_library *library)
{
	dlclose(library->handle);
}

/**
 *  OS dependent part to resolve a symbol
 */
void *gpmi_dlsym_lowlevel(gpmi_library *library, const char *name)
{
	return dlsym(library->handle, name);
}

/**
 *  OS dependent part to get a handle to the global address space
 */
void *gpmi_dlopen_global()
{
	return dlopen(NULL, RTLD_GLOBAL);
}

char *gpmi_getenv(gpmi_env_t env)
{

#ifdef UNIX
	switch(env) {
		case gpmi_env_home:
			return strdup(getenv("HOME"));
	}
#endif
#ifdef WIN32
	switch(env) {
		case gpmi_env_home: {
			char path[MAX_PATH];
			SHGetFolderPath(NULL, CSIDL_APPDATA|CSIDL_FLAG_CREATE, NULL, 0, path);
			return strdup(path);
		}
	}
#endif
	assert(!"Invalid env in gpmi_getenv or porting problem");
	return NULL;
}
