/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#define _GNU_SOURCE 1 // Needed for recursive thread-locking
#ifdef gpmi_threads
#include <pthread.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "threads.h"


/* These globals are written only before switcing to threading so they are safe */
gpmi_shared_globals_t  gpmi_shared;
gpmi_private_globals_t gpmi_private;
int gpmi_threading                  = 0;
static int gpmi_threads_initialized = 0;

void gpmi_threads_init_globals(gpmi_shared_globals_t *sg, gpmi_private_globals_t *pg)
{
	if (sg != NULL) {
		sg->gpmi_keywords         = NULL;
		sg->error_stack_enable    = 0;
		sg->gpmi_events           = NULL;
		sg->event_types           = NULL;
		sg->event_types_allocated = 0;
		sg->gpmi_libraries        = NULL;
		sg->static_modules        = NULL;
		sg->static_modules_count  = 0;
		sg->static_packages       = NULL;

		sg->path_packages = NULL;
		sg->path_modules  = NULL;
		sg->path_scripts  = NULL;
		memset(&(sg->main_lib), 0, sizeof(sg->main_lib));
	}

	if (pg != NULL) {
		#ifdef gpmi_malloc_cache
			pg->global_argv  = NULL;
			pg->global_argvn = 0;
		#endif
		pg->err_stack      = NULL;
		pg->current_module = NULL;

		/* XXX: should be #ifdef gpmi_malloc_cache */
		pg->pkg_misc_func_buff = NULL;
		pg->pkg_misc_func_size = 0;
	}
}

void gpmi_threads_free_private(gpmi_private_globals_t *pg)
{
	/* Free the private area for the thread */
	#ifdef gpmi_malloc_cache
		if (pg->global_argv != NULL)
			free(pg->global_argv);
	#endif
	if (pg->pkg_misc_func_buff != NULL)
		free(pg->pkg_misc_func_buff);
	if (pg->err_stack != NULL)
		gpmi_err_stack_destroy(NULL);
}


#ifdef gpmi_threads
pthread_mutex_t gpmi_shared_mutex = PTHREAD_RECURSIVE_MUTEX_INITIALIZER_NP;
pthread_key_t   gpmi_private_key;


/* Frees all fields and the pointer itself as well */
void gpmi_threads_free_private_ptr(gpmi_private_globals_t *pg)
{
	gpmi_threads_free_private(pg);
	free(pg);
}


int gpmi_threads_init(void)
{
/* IMPORTANT: the init part must not depend on the value if gpmi_threading,
because it may be unknown. Static package initializers call gpmi_init() before
the app, so the first init is always non-threading. */
	if (gpmi_threads_initialized > 0)
		return 0;

	gpmi_threads_init_globals(&gpmi_shared, &gpmi_private);

	gpmi_threads_initialized++;
	return 0;
}

int gpmi_threads_uninit(void)
{

	if (gpmi_threads_initialized > 0)
		gpmi_threads_free_private(&gpmi_private);

	return 0;
}

gpmi_private_globals_t *gpmi_threads_get_private(void)
{
	gpmi_private_globals_t *priv;

	if (gpmi_threading == 0)
		return &gpmi_private;

	priv = pthread_getspecific(gpmi_private_key);
	if (priv != NULL)
		return priv;

	priv = malloc(sizeof(gpmi_private_globals_t));
	gpmi_threads_init_globals(NULL, priv);
	assert(pthread_setspecific(gpmi_private_key, priv) == 0);

	return priv;
}

void gpmi_threads_start(void)
{
	gpmi_threading = 1;
	pthread_key_create(&gpmi_private_key, (void *)gpmi_threads_free_private_ptr);
}

#else
int gpmi_threads_init(void)
{
	assert(gpmi_threading == 0);

	if (gpmi_threads_initialized > 0)
		return 0;

	/* Initialize the two global variables */
	gpmi_threads_init_globals(&gpmi_shared, &gpmi_private);

	gpmi_threads_initialized++;
	return 0;
}

int gpmi_threads_uninit(void)
{
	assert(gpmi_threading == 0);
	/* No need to free static global vars */

	gpmi_threads_free_private(&gpmi_private);

	return 0;
}

gpmi_private_globals_t *gpmi_threads_get_private(void)
{
	/* In single thread case we have only one storage for private globals */
	return &gpmi_private;
}

void gpmi_threads_start(void)
{
}
#endif
