/*
    libgpmi - General Package/Module Interface
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef GPMI_THREADS_H
#define GPMI_THREADS_H

#ifdef gpmi_threads
#include <pthread.h>
#endif


/*

Threads.[ch] is used even if GPMI compiled without threads: it gives a
general API so that no other part of GPMI needs to know too much about threading
or not-threading. The basic idea is to have 2 big structures, one holding shared
global variables (all threads share them), the other contains all the
private globals (per thread basis). The shared one has a locking mechanism,
before changing anything in it, the caller needs to lock, after the write
unlock.

In a non-threading compile both are simple global structures and locking
doesn't do anyting (empty macros).
*/

#include <gpmi/keywords.h>
#include <gpmi/events.h>
#include <gpmi/error.h>
#include <gpmi/hash/hash.h>

/* This structure holds all the global variables that need to be shared among
threads. This way we can use a single lock and, reducing the chance of a
deadlock situation. In non-threading version, we just define a global variable
of this structure. */

typedef struct gpmi_shared_globals_s {
	gpmi_keyword *gpmi_keywords;     //!< list of all keyword tables
	gpmi_keyword *gpmi_events;       //!< list of all events
	char **event_types;              //!< Here we store the parameter types of all events here
	int event_types_allocated;       //!< Here we keep track of the allocated event_types

	int error_stack_enable;          //!< Controlling the error stacck: 0 means disabled, 1 means enabled
	gpmi_package *dynamic_packages;  //!< Linked list of dynamic packages which the user didn't want to have a pointer for
	gpmi_library *gpmi_libraries;    //!< double linked list of libraries
	hash_t *static_modules;          //!< list of all static modules
	int static_modules_count;        //!< number of active static modules
	gpmi_package *static_packages;   //!< list of static packages

	/* Search paths */
	char *path_modules;
	char *path_packages;
	char *path_scripts;

	gpmi_library main_lib;           //!< Fake library which holds the dependency of the application
} gpmi_shared_globals_t;

extern gpmi_shared_globals_t gpmi_shared;

#ifdef gpmi_threads
	/* Locking mechanism for gpmi_shared using mutexes */
	extern pthread_mutex_t gpmi_shared_mutex;
	#ifdef gpmi_threads_debug
		#define gpmi_shared_lock       gpmi_threading == 0 ? 0 : pthread_mutex_lock(&gpmi_shared_mutex) ; printf("[threads-debug]   lock at %s:%d (%s)\n", __FILE__, __LINE__, __FUNCTION__)
		#define gpmi_shared_unlock     gpmi_threading == 0 ? 0 : pthread_mutex_unlock(&gpmi_shared_mutex) ; printf("[threads-debug] unlock at %s:%d (%s)\n", __FILE__, __LINE__, __FUNCTION__)
	#else
		#define gpmi_shared_lock       gpmi_threading == 0 ? 0 : pthread_mutex_lock(&gpmi_shared_mutex)
		#define gpmi_shared_unlock     gpmi_threading == 0 ? 0 : pthread_mutex_unlock(&gpmi_shared_mutex)
	#endif
#else
	/* We don't need to lock in no-thread-mode */
	#define gpmi_shared_lock
	#define gpmi_shared_unlock
#endif

/* This structure holds all the global variables that should be used as local
storages. Most common use of these are persistent buffers to reduce the number
of malloc calls. In non-threading version, we define a global var from this
structure. In both threading and non-threading versions functions will use
a common gateway proc to learn the address of the structure. */

typedef struct gpmi_private_globals_s {
	#ifdef gpmi_malloc_cache
		gpmi_globev_param_t *global_argv;
		int global_argvn;
	#endif
	gpmi_err_stack_t *err_stack;      //<! Error stack of the current thread
	gpmi_module *current_module;      //<! link to the current module active
	char *pkg_misc_func_buff;         //<! XXX: should be #ifdef gpmi_malloc_cache
	int pkg_misc_func_size;           //<! XXX: should be #ifdef gpmi_malloc_cache
} gpmi_private_globals_t;

extern gpmi_private_globals_t gpmi_private;

gpmi_private_globals_t *gpmi_threads_get_private(void); //!< returns a pointer to the private data of the thread

int gpmi_threads_init(void);         //!< initializes threading API (for internal use)
int gpmi_threads_uninit(void);       //!< uninitializes threading API (for internal use)
void gpmi_threads_start(void);       //!< this tells gpmi that the application starts threading. Call it before creating the first thread.

/* Internals - do not use unless you are sure... */
extern int gpmi_threading;

#endif /* GPMI_THREADS_H */
