/*
    libgpmi - General Package/Module Interface - python module
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "python.h"

#include <stdio.h>
#include <string.h>
#ifdef WIN32
/* For basename and dirname on windows: */
#include "gpmi/os_dep.h"
#else
#include <libgen.h>
#endif

#include "gpmi/debug.h"

static int python_module_loaded = 0;

static PyObject *Python_DynamicPackageCommand(PyObject *self, PyObject *args)
{
	char *(*proc)(int argc, char *argv[]);
	PyObject *py_res;
	char **new_args;
	char *res;
	int argc, i;
	char *function_name = (char *)self;

	/* Find function-proc */
	proc = GPMI_Script_FindProc(function_name);

	/* Safe checking */
	if (proc == NULL) {
		gpmi_error(0, "[Python] Internal error: %s has a NULL proc in the command hash\n", function_name);
		return Py_None;
	}

	/* Check params */
	argc = PyTuple_Size(args);

	/* Allocate memory for the args */
	new_args = malloc((argc + 1) * sizeof(char *));

	/* Set the first param */
	new_args[0] = function_name;

	/* Convert all params to a string */
	for (i = 0; i < argc; i++) {
		PyObject *name = PyTuple_GetItem(args, i);

		/* Handle all types... */
		new_args[i + 1] = strdup(PyString_AsString(PyObject_Str(name)));
	}

	/* Call the real function */
	res = proc(argc + 1, new_args);

	/* Free the strings */
	for (i = 0; i < argc; i++) {
		free(new_args[i + 1]);
	}
	free(new_args);

	/* If res is NULL, return nothing */
	if (res == NULL) {
		return Py_None;
	}

	/* Else return a string */
	py_res = PyString_FromString(res);
	free(res);

	/* Return nothing */
	return py_res;
}

/* Array needed to register commands */
static PyMethodDef python_empty_method[] = {
	{"Test", Python_DynamicPackageCommand, METH_VARARGS, NULL},
	{NULL, NULL, 0, NULL}
};

void Python_RegisterCommand(void *client_data, const char *name)
{
	Python_ClientData *Python_MyData = client_data;
	PyObject *new;
	PyMethodDef *new_func_define;
	char *new_name = strdup(name);

	/* Assign the new name */
	python_empty_method->ml_name = new_name;

	new_func_define = malloc(sizeof(python_empty_method));
	memcpy(new_func_define, python_empty_method, sizeof(python_empty_method));

	/* Register it */
	new = PyCFunction_New(new_func_define, NULL);
	/* Here we misuse m_self to assing a 'self' to the function. This way we can
	 *  later on recover the functionname in an easy way. */
	((PyCFunctionObject *)new)->m_self = (PyObject *)new_name;

	/* If not NULL, assign the name */
	if (new == NULL || PyDict_SetItemString(Python_MyData->root_dict, new_name, new)) {
		gpmi_error(0, "[Python] Could not register %s\n", name);

		/* XXX -- Ugly way to print errors */
		if (PyErr_Occurred())
			PyErr_Print();

		return;
	}

	/* Free some temp stuff */
	Py_DECREF(new);
}

int Python_LoadScript(void *client_data, const char *filename)
{
	Python_ClientData *Python_MyData = client_data;
	char path[512], file[512], temp[512];
	char *rfile, *rpath;

	/* Asking what we do here is most likely not a good idea... okay, some information:
	 *  Python doesn't allow dir-changing via 'import' command (which we do use here).
	 *  So what we do.. we add the path to the searchpath, then include the file..
	 *  Stupid? Yes! Does it work? That too ;) */
	snprintf(path, sizeof(path), "%s", filename);
	snprintf(file, sizeof(file), "%s", filename);

	rpath = dirname(path);
	rfile = basename(file);

	snprintf(temp, sizeof(temp), "sys.path.append('%s')", rpath);

	PyRun_SimpleString(temp);

	/* Load the module */
	Python_MyData->module = PyImport_ImportModule(rfile);

	if (Python_MyData->module == NULL) {
		/* We did not load the file? */
		gpmi_error(0, "[Python] Could not load file %s\n", filename);

		/* XXX -- Ugly way to print errors */
		if (PyErr_Occurred())
			PyErr_Print();

		return -1;
	}

	Python_MyData->dict = PyModule_GetDict(Python_MyData->module);

	return 0;
}

void Python_CallFunction(void *client_data, const char *function, int argc, const char **argv)
{
	Python_ClientData *Python_MyData = client_data;
	PyObject *pFunc, *pArgs, *pValue;
	int i;

	pFunc = PyDict_GetItemString(Python_MyData->dict, function);

	if (pFunc == NULL || !PyCallable_Check(pFunc)) {
		gpmi_error(0, "[Python] Could not launch function %s\n", function);

		/* XXX -- Ugly way to print errors */
		if (PyErr_Occurred())
			PyErr_Print();

		return;
	}

	/* Create the args-list */
	pArgs = PyTuple_New(argc);

	for (i = 0; i < argc; i++ ) {
		pValue = PyString_FromString(argv[i]);
		PyTuple_SetItem(pArgs, i, pValue);
	}

	pValue = PyObject_CallObject(pFunc, pArgs);

	Py_DECREF(pArgs);

	if (pValue != NULL)
		{ Py_DECREF(pValue); } // Bad macros!

	if (PyErr_Occurred())
		PyErr_Print();
}

int Python_Initialize(void **client_data)
{
	Python_ClientData *MyData;

	if (python_module_loaded >= 1) {
		gpmi_error(0, "[Python] This module can only have one active instance");
		return -1;
	}

	MyData = malloc(sizeof(Python_ClientData));
	memset(MyData, 0, sizeof(Python_ClientData));

	*client_data = MyData;

	Py_Initialize();

	PyRun_SimpleString("import sys");

	MyData->root_module = Py_InitModule("__builtin__", python_empty_method);
	MyData->root_dict = PyModule_GetDict(MyData->root_module);

	python_module_loaded++;

	return 0;
}

void Python_Destroy(void **client_data)
{
	Python_ClientData *Python_MyData = *client_data;

	/* Free all our own temporary shit */
	{
		PyCFunctionObject *func;
		char *tmp_funcname;
		PyMethodDef *ml;
		PyDictObject *mp;
		PyDictEntry *ep, *table;
		int fill;

		/* Get our dict */
		mp = (PyDictObject *)(Python_MyData->root_dict);
		table = mp->ma_table;
		fill = mp->ma_fill;

		/* Go through all dict-items */
		for (ep = table; fill > 0; ep++) {
			if (ep->me_key == NULL)
				continue;

			fill--;

			/* Is it a valid function */
			if (!PyCallable_Check(ep->me_value))
				continue;

			func = (PyCFunctionObject *)(ep->me_value);
			ml = func->m_ml;

			/* Is it a function we added? */
			if (ml == NULL || func->m_self == NULL || ml->ml_meth != Python_DynamicPackageCommand)
				continue;

			tmp_funcname = (char *)func->m_self;
			func->m_self = NULL;
			/* Free the function */
			PyDict_DelItemString(Python_MyData->root_dict, tmp_funcname);

			/* Free the function-name and ml which was stored here */
			free(tmp_funcname);
			free(ml);
		}
	}

	Py_Finalize();

	if (Python_MyData->module != NULL) {
		Py_DECREF(Python_MyData->module);
		Py_DECREF(Python_MyData->root_module);
	}

	/* Finally free up the client-data */
	free(*client_data);

	python_module_loaded--;
}

char *Python_GetExtension()
{
	/* Python doesn't like extensions.... :s */
	return strdup("");
}
