/*
    libgpmi - General Package/Module Interface - tcl module
    Copyright (C) 2005-2007 Patric 'TrueLight' Stout & Tibor 'Igor2' Palinkas
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "tcl.h"

#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "gpmi/debug.h"

/**
 * Returns the extension of this script-module.
 */
char *TCL_GetExtension()
{
	return strdup(".tcl");
}

/**
 * Reloads the script on the fly.
 */
int TCL_Rehash(void *client_data)
{
	TCL_ClientData *TCL_MyData = client_data;

	return Tcl_EvalFile(TCL_MyData->interp, TCL_MyData->script_name);
}

/**
 * Loads a tcl script.
 */
int TCL_LoadScript(void *client_data, const char *filename)
{
	TCL_ClientData *TCL_MyData = client_data;

	/* Save the (new?) filename of the script - for rehash */
	if (TCL_MyData->script_name != NULL )
		free(TCL_MyData->script_name);

	TCL_MyData->script_name = strdup(filename);

	/* Read the file */
	if (Tcl_EvalFile(TCL_MyData->interp, filename) == TCL_OK) {
		return 0;
	} else {
		TCL_ReportTclError(TCL_MyData);
		return -1;
	}
}

/**
 * Reports a tcl error using DEBUG. Call this if you get something else
 * than TCL_OK.
 */
void TCL_ReportTclError(TCL_ClientData *TCL_MyData)
{
	Tcl_Obj *res;

	res = Tcl_GetObjResult(TCL_MyData->interp);
	gpmi_error(0, "[TCL] Tcl error: %s\n", res->bytes);
}

/**
 * Execute a dynamic command.
 **/
int TCL_DynamicPackageCommand(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
	TCL_ClientData *TCL_MyData = GPMI_Script_GetClientData();
	char *(*proc)(int argc, char *argv[]);
	char *s;

	proc = GPMI_Script_FindProc(argv[0]);

	/* Check if command not found - this must be an internal error: the
	   package had registered the command in the hash, but the routine
	   pointer is NULL.
	*/
	if (proc == NULL) {
		gpmi_error(0, "[TCL] Internal error: %s has a NULL proc in the command hash\n", argv[0]);
		return TCL_OK;
	}

	/* Run command, append string result to Tcl result */
	s = proc(argc, argv);

	if (s != NULL) {
		Tcl_AppendResult(TCL_MyData->interp, s, NULL);

		/* Free result */
		free(s);
	}

	return TCL_OK;
}

/**
 * Register a command in our command hash
 */
void TCL_RegisterCommand(void *client_data, const char *name)
{
	TCL_ClientData *TCL_MyData = client_data;

	Tcl_CreateCommand(TCL_MyData->interp, name, (Tcl_CmdProc *)TCL_DynamicPackageCommand, (ClientData)NULL, (Tcl_CmdDeleteProc *)NULL);
}

/**
 * Call a function of the script by name, passing params.
 */
void TCL_RunFunction(void *client_data, const char *function, int argc, const char **argv)
{
	TCL_ClientData *TCL_MyData = client_data;
	char *func_param, *s;

	s = Tcl_Merge(argc, argv);

	func_param = malloc(strlen(s) + strlen(function) + 32);
	sprintf(func_param, "%s %s", function, s);

	Tcl_Free(s);

	if (Tcl_Eval(TCL_MyData->interp, func_param) != TCL_OK)
		TCL_ReportTclError(TCL_MyData);

	free(func_param);
}

int TCL_Initialize(void **client_data)
{
	TCL_ClientData *MyData;

	MyData = malloc(sizeof(TCL_ClientData));
	memset(MyData, 0, sizeof(TCL_ClientData));

	*client_data = MyData;

	Tcl_FindExecutable("");

	/* Initialize a tcl interpreter */
	MyData->interp = Tcl_CreateInterp();

	if (Tcl_Init(MyData->interp) == TCL_ERROR) {
		free(MyData);
		return -1;
	}

	return 0;
}

void TCL_Destroy(void **client_data)
{
	TCL_ClientData *TCL_MyData = *client_data;

	/* Not initialized */
	if (TCL_MyData->script_name == NULL)
		return;

	/* Free the script */
	free(TCL_MyData->script_name);
	TCL_MyData->script_name = NULL;

	Tcl_DeleteInterp(TCL_MyData->interp);

	/* Finally free up the client-data */
	free(*client_data);
}
